// False positive with itertools::process_results
#![allow (clippy::redundant_closure)]

use std::{
	collections::HashMap,
	convert::{TryFrom},
	iter::FromIterator,
	path::Path,
};

use crate::errors::ConfigError;

// Stuff we need to load from the config file and use to
// set up the HTTP server

pub mod file {
	use serde::Deserialize;
	
	use crate::key_validity::*;
	
	#[derive (Deserialize)]
	pub struct Server {
		// This is duplicated in the hashmap, but it's not a problem
		pub name: String,
		pub tripcode: BlakeHashWrapper,
		pub display_name: Option <String>,
	}
	
	#[derive (Deserialize)]
	pub struct DevMode {
		pub scraper_key: Option <ScraperKey <Valid7Days>>,
	}
	
	// Stuff that's identical between the file and the runtime structures

	#[derive (Default, Deserialize)]
	pub struct Isomorphic {
		#[serde (default)]
		pub enable_scraper_auth: bool,
		
		// If any of these fields are used, we are in dev mode and have to 
		// show extra warnings, since some auth may be weakened
		pub dev_mode: Option <DevMode>,
	}
	
	#[derive (Deserialize)]
	pub struct Config {
		pub port: Option <u16>,
		pub servers: Vec <Server>,
		#[serde (flatten)]
		pub iso: Isomorphic,
	}
}

// Stuff we actually need at runtime

pub struct Config {
	pub port: Option <u16>,
	pub servers: HashMap <String, file::Server>,
	pub iso: file::Isomorphic,
}

impl TryFrom <file::Config> for Config {
	type Error = ConfigError;
	
	fn try_from (f: file::Config) -> Result <Self, Self::Error> {
		let servers = f.servers.into_iter ()
		.map (|server| Ok::<_, ConfigError> ((server.name.clone (), server)));
		
		let servers = itertools::process_results (servers, |i| HashMap::from_iter (i))?;
		
		Ok (Self {
			port: f.port,
			servers,
			iso: f.iso,
		})
	}
}

impl Config {
	pub async fn from_file (path: &Path) -> Result <Self, ConfigError> {
		use tokio::prelude::*;
		
		let mut f = tokio::fs::File::open (path).await?;
		
		let mut buffer = vec! [0_u8; 4096];
		let bytes_read = f.read (&mut buffer).await?;
		buffer.truncate (bytes_read);
		
		let config_s = String::from_utf8 (buffer)?;
		let new_config: file::Config = toml::from_str (&config_s)?;
		
		Self::try_from (new_config)
	}
}
